from __future__ import absolute_import

import json
import os
import pytest
import requests
import time

from rancher_gen.compat import b64encode


@pytest.fixture(scope='session')
def stack_service(request):
    host = os.getenv('RANCHER_HOST')
    port = int(os.getenv('RANCHER_PORT', 80))
    access_key = os.getenv('RANCHER_ACCESS_KEY')
    secret_key = os.getenv('RANCHER_SECRET_KEY')
    api_token = b64encode("{0}:{1}".format(access_key, secret_key))

    with open(os.path.join(os.path.dirname(__file__), 'compose.yaml')) as fh:
        compose_string = fh.read()

    # Create the stack and service
    headers = {
        'Authorization': 'Basic {0}'.format(api_token)
    }
    url = 'http://{0}:{1}/v1/environments'.format(host, port)
    stack_name = 'teststack'
    res = requests.post(url, headers=headers, data={
        'name': stack_name,
        "dockerCompose": compose_string,
        "startOnCreate": True
    })
    stack = res.json()

    # Wait for stack to be active
    state = stack['state']
    url = '{0}?name={1}'.format(url, stack_name)
    while state != 'active':
        res = requests.get(url, headers=headers)
        stack = res.json()['data'][0]
        state = stack['state']
        time.sleep(1)

    # Wait for services to be active
    services = []
    for i in [1, 2]:
        url = '{0}?name={1}'.format(stack['links']['services'], 'hello%d' % i)
        res = requests.get(url, headers=headers)
        state = ''
        service = None
        while state != 'active':
            service = res.json()['data'][0]
            state = service['state']
            time.sleep(1)
        services.append(service)

    def teardown():
        requests.delete('{0}/{1}'.format(url, stack['id']), headers=headers)
    request.addfinalizer(teardown)

    return stack, services


@pytest.fixture(scope='function')
def mock_message(request, stack_service):
    stack, services = stack_service
    service_id = services[0]['id']
    with open(os.path.join(os.path.dirname(__file__), 'mock_msg.json')) as fh:
        mock_message = json.loads(fh.read())
    instances_link =\
        mock_message['data']['resource']['services'][0]['links']['instances']
    mock_message['data']['resource']['services'][0]['links']['instances'] =\
        instances_link.format(stack['accountId'], services[0]['id'])

    return mock_message



from __future__ import absolute_import

import json
import logging
import os
import requests
import ssl
import websocket

from jinja2 import Environment, FileSystemLoader
from subprocess import call
from threading import Thread

from .compat import b64encode
from .rancher import API

logger = logging.getLogger(__name__)

# Change the log level on the requests library
logging.getLogger("requests").setLevel(logging.WARNING)


class RancherConnector(object):

    def __init__(self, host, port, project_id, access_key, secret_key,
                 template, dest, ssl=False, stack=None, services=None,
                 notify=None):
        self.rancher_host = host
        self.rancher_port = port
        self.project_id = project_id
        self.api_token = b64encode("{0}:{1}".format(access_key, secret_key))
        self.template = template
        self.dest = dest
        self.ssl = ssl
        self.stack = stack
        self.services = services
        self.notify = notify

    def __call__(self):
        self._prerender()
        self.start()

    def _prerender(self):
        instances = None
        api = API(self.rancher_host, self.rancher_port, self.project_id,
                  self.api_token, self.ssl)

        # If we're not filtering by stack and services, then load all instances
        # in the environment
        if self.stack is None and self.services is None:
            instances = api.get_instances()

        # If we're only filterting by stack, then load the instances for that
        # stack
        elif self.stack and self.services is None:
            instances = api.get_instances(stack_name=self.stack)

        # If we're filtering by stack and service, then load the instances for
        # that service
        elif self.stack and self.services and len(self.services) > 0:
            api = API(self.rancher_host, self.rancher_port, self.project_id,
                      self.api_token, self.ssl)
            services = api.get_services(self.stack, self.services)

            instances = []
            if len(services) > 0:
                for service in services:
                    instances += api.get_instances(service)

        # If we found any instnaces, then render the template
        if instances is None:
            instances = []
        render_template(instances, self.template, self.dest)
        notify(self.notify)

    def start(self):
        header = {
            'Authorization': 'Basic {0}'.format(self.api_token)
        }
        protocol = 'wss' if self.ssl else 'ws'
        url = '{0}://{1}:{2}/v1/projects/{3}/subscribe?eventNames='\
            'resource.change&include=services'\
            .format(protocol, self.rancher_host, self.rancher_port,
                    self.project_id)
        self.ws = websocket.WebSocketApp(url, header=header,
                                         on_message=self._on_message,
                                         on_open=self._on_open,
                                         on_error=self._on_error,
                                         on_close=self._on_close)

        logger.info('Watching for rancher events')
        self.ws.run_forever()

    def _on_open(self, ws):  # pragma: no cover
        logger.info("Websocket connection open")

    def _on_close(self, ws):  # pragma: no cover
        logger.info('Websocket connection closed')

    def _on_error(self, ws, error):  # pragma: no cover
        logger.error(error)

    def _on_message(self, ws, message):
        msg = json.loads(message)
        if msg['name'] == 'resource.change' and msg['data']:
            handler = MessageHandler(msg, self.rancher_host, self.rancher_port,
                                     self.project_id, self.api_token,
                                     self.template, self.dest, self.ssl,
                                     self.stack, self.services, self.notify)
            handler.start()


class MessageHandler(Thread):
    def __init__(self, message, host, port, project_id, api_token, template,
                 dest, ssl, stack=None, services=None, notify=None):
        Thread.__init__(self)
        self.message = message
        self.rancher_host = host
        self.rancher_port = port
        self.project_id = project_id
        self.api_token = api_token
        self.template = template
        self.dest = dest
        self.ssl = ssl
        self.stack = stack
        self.services = services
        self.notify = notify

    def run(self):
        resource = self.message['data']['resource']
        if resource['type'] == 'container' and \
                resource['state'] in ['running', 'removed', 'stopped']:

            api = API(self.rancher_host, self.rancher_port, self.project_id,
                      self.api_token, self.ssl)

            # Filter by stack and/or service name if specified
            if self.stack:
                # Some instnaces like Network Agents don't have labels, and
                # don't really belong to a stack and/or service, so if this
                # message has to do with one of those instances, then we
                # need to ignore it.
                if resource['labels'] is None:
                    return

                stack_name = resource['labels']['io.rancher.stack.name']
                service_name =\
                    resource['labels']['io.rancher.stack_service.name']
                service_name = service_name.split('/')[1]

                # if the stacks don't match, then simply return
                if stack_name != self.stack:
                    return

                # If we're filtering by service, then load the instances for
                # this sevice
                if self.services and len(self.services) > 0:
                    if service_name not in self.services:
                        return

                    services = api.get_services(self.stack, self.services)
                    instances = []
                    if len(services) > 0:
                        for service in services:
                            instances += api.get_instances(service)
                        self._render_and_notify(instances)
                        return

                instances = api.get_instances(stack_name=self.stack)
                self._render_and_notify(instances)
            else:
                instances = api.get_instances()
                self._render_and_notify(instances)

    def _render_and_notify(self, instances):
        if instances is None:
            render_template([], self.template, self.dest)
        else:
            render_template(instances, self.template, self.dest)
        notify(self.notify)


def render_template(instances, template, dest):
    template_dir = os.path.dirname(template)
    template_filename = os.path.basename(template)
    env = Environment(loader=FileSystemLoader(template_dir))
    template = env.get_template(template_filename)
    result = template.render(containers=instances)

    with open(dest, 'w') as fh:
        fh.write(result)
    logger.info("Generated '{0}'".format(dest))


def notify(notify):
    if notify:
        logger.info("Running '{0}'".format(notify))
        call(notify, shell=True)



from __future__ import absolute_import
import requests


class API(object):

    def __init__(self, host, port, project_id, api_token, ssl):
        self.host = host
        self.port = port
        self.project_id = project_id
        self.api_token = api_token
        self.ssl = ssl
        self._headers = {
            'Authorization': 'Basic {0}'.format(api_token)
        }
        self._protocol = 'https' if ssl else 'http'

    def get_services(self, stack, services):
        _services = []
        for service in services:
            _services.append(self._get_service_from_stack(stack, service))

        return _services

    def get_service(self, resource=None, stack=None, service=None):
        """ Get a service from a rancher resource or by stack and name.

        If resource is specified, stack and service and ignore. If resource is
        None, then stack and service are used to lookup the service in Rancher.

        Args:
            - resource: A rancher reource the represents a message sent by the
                web socket.
            - stack: The name of the stack
            - service: The name of the service
        """
        if resource:
            return self._get_service_from_resource(resource)
        elif stack and service:
            return self._get_service_from_stack(stack, service)
        return None

    def get_instances(self, service=None, stack_name=None):
        """ Gets the list of instances from a service.

        If service is specified, it's corresponding links are used to look up
        the intances. If stack_name is specified, it is used used load all
        instances that belong to that stack. If neither service nor stack_name
        are specified, all instances in the environment are returned.

        Args:
            - service: The Rancher object representation of a service
            - stack_name: A stack name
        """
        if service is None:
            url = '{0}://{1}:{2}/v1/projects/{3}/instances'\
                .format(self._protocol, self.host, self.port, self.project_id)
        else:
            url = service['links']['instances']
        res = requests.get(url, headers=self._headers)
        res_data = res.json()
        if res_data['data']:
            if stack_name is None:
                return res_data['data']

            instances = []
            for resource in res_data['data']:
                labels = resource['labels']
                if labels and 'io.rancher.stack.name' in labels and \
                        labels['io.rancher.stack.name'] == stack_name:
                    instances.append(resource)
            if len(instances) > 0:
                return instances

        return None

    def _get_service_from_resource(self, resource):
        # If the state is running or stopped, then the list of services is
        # already part of the resource, so simply return the first service.
        if resource['state'] in ['running', 'stopped']:
            return resource['services'][0]
        elif resource['state'] == 'removed':
            # If a container was removed, we need to search the service by
            # the stack name
            stack_name = resource['labels']['io.rancher.stack.name']
            service_name = resource['labels']['io.rancher.stack_service.name']
            service_name = service_name.split('/')[1]
            return self._get_service_from_stack(stack_name, service_name)
        return None

    def _get_service_from_stack(self, stack_name, service_name):
        # First get the stack
        url = '{0}://{1}:{2}/v1/projects/{3}/environments'\
            .format(self._protocol, self.host, self.port, self.project_id)
        res = requests.get(url, headers=self._headers)
        res_data = res.json()
        service_stack = None
        if res_data['data'] and len(res_data['data']) > 0:
            for stack in res_data['data']:
                if stack['name'] == stack_name:
                    service_stack = stack
                    break

        if not service_stack:
            return None

        url = service_stack['links']['services']
        res = requests.get(url, headers=self._headers)
        res_data = res.json()
        if res_data['data'] and len(res_data['data']) > 0:
            for service in res_data['data']:
                if service['name'] == service_name:
                    return service

        return None



from __future__ import absolute_import, print_function

import logging
import sys
from argparse import ArgumentParser, Action

from .handler import RancherConnector

logger = logging.getLogger(__name__)

LOG_LEVELS = {
    'DEBUG': logging.DEBUG,
    'INFO': logging.INFO,
    'WARNING': logging.WARNING,
    'ERROR': logging.ERROR,
    'CRITICAL': logging.CRITICAL
}


class SetLogLevel(Action):
    def __call__(self, parser, namespace, values, option_string=None):
        level = values.upper()
        if level not in LOG_LEVELS:
            raise ValueError("Invalid log level: {0}. Must be one of {1}"
                             .format(values, LOG_LEVELS))
        logging.getLogger('rancher_gen').setLevel(LOG_LEVELS[level])


def validate_args(args):
    if not args.host:
        print ("error: Missing host parameter")
        return False

    if not args.access_key:
        print ("error: Missing Rancher access key parameter")
        return False

    if not args.secret_key:
        print ("error: Missing Rancher secret key parameter")
        return False

    if not args.project_id:
        print ("error: Missing Rancher project id parameter")
        return False

    return True


def main():
    parser = ArgumentParser('rancher-gen', add_help=False,
                            description="Generate files from rancher meta-data")

    named_args = parser.add_argument_group('named arguments')
    named_args.add_argument('--host',
                            help='Rancher host (defaults to localhost)')
    named_args.add_argument('--port', type=int, default=-1,
                            help='Rancher port (defaults to 80 or 443 if ssl '
                            'is specified)')
    named_args.add_argument('--access-key', help='The Rancher access key')
    named_args.add_argument('--secret-key', help='The Rancher secret key')
    named_args.add_argument('--project-id', help="Rancher's project id")

    optional_args = parser.add_argument_group('optional arguments')
    optional_args.add_argument("-h", "--help", action="help",
                               help="show this help message and exit")
    optional_args.add_argument('--log-level', action=SetLogLevel,
                               default='INFO', choices=LOG_LEVELS,
                               help='Set the log level.')
    optional_args.add_argument('--stack',
                               help="The name of the rancher stack")
    optional_args.add_argument('--service', action="append", metavar="SERVICE",
                               dest="services",
                               help="The name of the rancher service")
    optional_args.add_argument('--ssl', action='store_true',
                               default=False,
                               help='User secure connections')
    optional_args.add_argument('--notify',
                               help="Command to run after template is "
                               "generated (e.g restart some-service)")

    parser.add_argument('template', help="Path to template to generate")
    parser.add_argument('dest', help="Output path for generated file")

    args = parser.parse_args()
    if not validate_args(args):
        return

    try:
        port = args.port
        if args.port == -1:
            port = 443 if args.ssl else 80
        handler = RancherConnector(args.host, port, args.project_id,
                                   args.access_key, args.secret_key,
                                   args.template, args.dest, args.ssl,
                                   args.stack, args.services, args.notify)
        handler()
    except Exception as e:
        logger.exception(e)


if __name__ == "__main__":
    main()



"""
Compatibility layer and utilities.
"""
import base64
import sys

PY3 = sys.version_info[0] >= 3

# Python 3 does not have StringIO, we should use the io module instead
try:
    from StringIO import StringIO  # noqa
except ImportError:
    from io import StringIO  # noqa


def b64encode(bytes_or_str):
    input_bytes = bytes_or_str
    if PY3 and isinstance(bytes_or_str, str):
        input_bytes = bytes_or_str.encode('ascii')

    output_bytes = base64.b64encode(input_bytes)
    if PY3:
        return output_bytes.decode('ascii')
    return output_bytes
